@@ -18,6 +18,7 @@ import os
 import sys
 from dataclasses import dataclass
 from math import ceil
+import gzip
 
 from tqdm import tqdm
 
@@ -38,7 +39,7 @@ def get_arguments():
     projects directory where outputs go.
     """
     parser = argparse.ArgumentParser(
-        prog="baler.py",
+        prog="baler",
         description=(
             "Baler is a machine learning based compression tool for big data.\n\n"
             "Baler has three running modes:\n\n"
@@ -72,6 +73,12 @@ def get_arguments():
         "  2. If workspace exists but project does not, create project in workspace.\n"
         "  3. If workspace does not exist, create workspace directory and project.",
     )
+    parser.add_argument(
+        "--pytorchProfile", action="store_true", help="Enable PyTorch profiling"
+    )
+    parser.add_argument(
+        "--energyProfile", action="store_true", help="Enable Energy profiling"
+    )
     parser.add_argument(
         "--verbose", dest="verbose", action="store_true", help="Verbose mode"
     )
@@ -91,11 +98,22 @@ def get_arguments():
         config = Config
         importlib.import_module(config_path).set_config(config)
 
-    return config, args.mode, workspace_name, project_name, args.verbose
+    return (
+        config,
+        args.mode,
+        workspace_name,
+        project_name,
+        args.verbose,
+        args.pytorchProfile,
+        args.energyProfile,
+    )
 
 
 def create_new_project(
-    workspace_name: str, project_name: str, verbose: bool = False
+    workspace_name: str,
+    project_name: str,
+    verbose: bool = False,
+    base_path: str = "workspaces",
 ) -> None:
     """Creates a new project directory output subdirectories and config files within a workspace.
 
@@ -105,8 +123,6 @@ def create_new_project(
         verbose (bool, optional): Whether to print out the progress. Defaults to False.
     """
 
-    base_path = "workspaces"
-
     # Create full project path
     workspace_path = os.path.join(base_path, workspace_name)
     project_path = os.path.join(base_path, workspace_name, project_name)
@@ -264,7 +280,14 @@ def normalize(data, custom_norm):
     return data
 
 
-def process(input_path, custom_norm, test_size, apply_normalization):
+def process(
+    input_path,
+    custom_norm,
+    test_size,
+    apply_normalization,
+    convert_to_blocks,
+    verbose,
+):
     """Loads the input data into a ndarray, splits it into train/test splits and normalizes if chosen.
 
     Args:
@@ -279,6 +302,14 @@ def process(input_path, custom_norm, test_size, apply_normalization):
     loaded = np.load(input_path)
     data = loaded["data"]
 
+    if verbose:
+        print("Original Dataset Shape - ", data.shape)
+
+    original_shape = data.shape
+
+    if convert_to_blocks:
+        data = data_processing.convert_to_blocks_util(convert_to_blocks, data)
+
     normalization_features = data_processing.find_minmax(data)
     if apply_normalization:
         print("Normalizing the data...")
@@ -291,11 +322,7 @@ def process(input_path, custom_norm, test_size, apply_normalization):
             data, test_size=test_size, random_state=1
         )
 
-    return (
-        train_set,
-        test_set,
-        normalization_features,
-    )
+    return (train_set, test_set, normalization_features, original_shape)
 
 
 def renormalize(data, true_min_list, feature_range_list):
@@ -401,6 +428,37 @@ def get_device():
     return device
 
 
+def save_error_bounded_requirement(config, decoded_output, data_batch):
+    rms_pred_error = (
+        np.divide(
+            np.subtract(decoded_output, data_batch),
+            data_batch,
+        )
+        * 100
+    )
+
+    # Ignoring RMS Undefind Values because Ground Truth is Zero
+    rms_pred_error[
+        (rms_pred_error == np.inf)
+        | (rms_pred_error == -np.inf)
+        | (rms_pred_error == np.nan)
+    ] = 0.0
+    rms_pred_error_index = np.where(
+        abs(rms_pred_error) > config.error_bounded_requirement
+    )
+    rows_idx, col_idx = rms_pred_error_index
+    if len(rows_idx) > 0 and len(col_idx) > 0:
+        rms_pred_error_exceeding_error_bound = np.subtract(
+            decoded_output,
+            data_batch,
+            dtype=np.float16,
+        )
+        deltas = []
+        for i in range(len(rows_idx)):
+            deltas.append(rms_pred_error_exceeding_error_bound[rows_idx[i]][col_idx[i]])
+    return deltas, rms_pred_error_index
+
+
 def compress(model_path, config):
     """Function which performs the compression of the input file. In order to compress, you must have a dataset whose
     path is determined by `input_path` in the `config`. You also need a trained model from path `model_path`. The
@@ -421,6 +479,13 @@ def compress(model_path, config):
     # Loads the data and applies normalization if config.apply_normalization = True
     loaded = np.load(config.input_path)
     data_before = loaded["data"]
+    original_shape = data_before.shape
+
+    if hasattr(config, "convert_to_blocks") and config.convert_to_blocks:
+        data_before = data_processing.convert_to_blocks_util(
+            config.convert_to_blocks, data_before
+        )
+
     if config.apply_normalization:
         print("Normalizing...")
         data = normalize(data_before, config.custom_norm)
@@ -428,7 +493,7 @@ def compress(model_path, config):
         data = data_before
     number_of_columns = 0
     try:
-        print("compression ratio:", config.compression_ratio)
+        n_features = 0
         if config.data_dimension == 1:
             column_names = np.load(config.input_path)["names"]
             number_of_columns = len(column_names)
@@ -436,10 +501,16 @@ def compress(model_path, config):
                 number_of_columns / config.compression_ratio
             )
             config.number_of_columns = number_of_columns
+            n_features = number_of_columns
         elif config.data_dimension == 2:
-            data = data_before
-            number_of_rows = data.shape[1]
-            config.number_of_columns = data.shape[2]
+            if config.model_type == "dense":
+                number_of_rows = data.shape[1]
+                config.number_of_columns = data.shape[2]
+                n_features = number_of_rows * config.number_of_columns
+            else:
+                number_of_rows = original_shape[1]
+                config.number_of_columns = original_shape[2]
+                n_features = config.number_of_columns
             config.latent_space_size = ceil(
                 (number_of_rows * config.number_of_columns) / config.compression_ratio
             )
@@ -461,24 +532,15 @@ def compress(model_path, config):
     model = data_processing.load_model(
         model_object,
         model_path=model_path,
-        n_features=config.number_of_columns,
+        n_features=n_features,
         z_dim=config.latent_space_size,
     )
     model.eval()
 
-    # Give the encoding function the correct input as tensor
-    # if config.data_dimension == 2:
-    #     data_tensor = (
-    #         torch.from_numpy(data.astype("float32", casting="same_kind"))
-    #         .to(device)
-    #         .view(data.shape[0], 1, data.shape[1], data.shape[2])
-    #     )
-    # elif config.data_dimension == 1:
-    #     data_tensor = torch.from_numpy(data).to(device)
     if config.data_dimension == 2:
         if config.model_type == "convolutional" and config.model_name == "Conv_AE_3D":
             data_tensor = torch.tensor(data, dtype=torch.float32).view(
-                1, 1, data.shape[0], data.shape[1], data.shape[2]
+                data.shape[0] // bs, 1, bs, data.shape[1], data.shape[2]
             )
         elif config.model_type == "convolutional":
             data_tensor = torch.tensor(data, dtype=torch.float32).view(
@@ -500,23 +562,59 @@ def compress(model_path, config):
     )
 
     # Perform compression
-    compressed = []
+    error_bound_batch, error_bound_deltas, error_bound_index, compressed = (
+        [],
+        [],
+        [],
+        [],
+    )
+
     with torch.no_grad():
         for idx, data_batch in enumerate(tqdm(data_dl)):
             data_batch = data_batch.to(device)
 
-            out = model.encode(data_batch)
+            compressed_output = model.encode(data_batch)
+
+            if config.save_error_bounded_deltas:
+                decoded_output = model.decode(compressed_output)
+                decoded_output = detacher(decoded_output)
+                deltas_compressed = 0
             # Converting back to numpyarray
-            out = detacher(out)
+            compressed_output = detacher(compressed_output)
+            data_batch = detacher(data_batch)
+
+            if config.save_error_bounded_deltas:
+                (
+                    deltas,
+                    rms_pred_error_index,
+                ) = save_error_bounded_requirement(config, decoded_output, data_batch)
+                if len(rms_pred_error_index) > 0:
+                    error_bound_batch.append(idx)
+                    error_bound_deltas.append(deltas)
+                    error_bound_index.append(rms_pred_error_index)
+                    deltas_compressed += len(rms_pred_error_index[0])
+
             if idx == 0:
-                compressed = out
+                compressed = compressed_output
             else:
-                compressed = np.concatenate((compressed, out))
+                compressed = np.concatenate((compressed, compressed_output))
 
-    return compressed
+    if config.save_error_bounded_deltas:
+        print("Total Deltas Found - ", deltas_compressed)
+
+    return (compressed, error_bound_batch, error_bound_deltas, error_bound_index)
 
 
-def decompress(model_path, input_path, model_name, config):
+def decompress(
+    model_path,
+    input_path,
+    input_path_deltas,
+    input_batch_index,
+    model_name,
+    config,
+    output_path,
+    original_shape,
+):
     """Function which performs the decompression of the compressed file. In order to decompress, you must have a
     compressed file, whose path is determined by `input_path`, a model from path `model_path` and a model_name. The
     model path and model names are used to initialize the model used for decompression. The data is then converted
@@ -537,12 +635,30 @@ def decompress(model_path, input_path, model_name, config):
     data = loaded["data"]
     names = loaded["names"]
     normalization_features = loaded["normalization_features"]
+
+    if config.model_type == "convolutional":
+        final_layer_details = np.load(
+            os.path.join(output_path, "training", "final_layer.npy"),
+        )
+
+    if config.save_error_bounded_deltas:
+        loaded_deltas = np.load(
+            gzip.GzipFile(input_path_deltas, "r"), allow_pickle=True
+        )
+        loaded_batch_indexes = np.load(
+            gzip.GzipFile(input_batch_index, "r"), allow_pickle=True
+        )
+        error_bound_batch = loaded_batch_indexes[0]
+        error_bound_deltas = loaded_deltas
+        error_bound_index = loaded_batch_indexes[1]
+        deltas_added = 0
+
     model_name = config.model_name
     latent_space_size = len(data[0])
     bs = config.batch_size
     model_dict = torch.load(str(model_path), map_location=get_device())
     if config.data_dimension == 2 and config.model_type == "dense":
-        number_of_columns = int(np.sqrt(len(model_dict[list(model_dict.keys())[-1]])))
+        number_of_columns = int((len(model_dict[list(model_dict.keys())[-1]])))
     else:
         number_of_columns = len(model_dict[list(model_dict.keys())[-1]])
 
@@ -557,6 +673,9 @@ def decompress(model_path, input_path, model_name, config):
     )
     model.eval()
 
+    if config.model_type == "convolutional":
+        model.set_final_layer_dims(final_layer_details)
+
     # Load the data, convert to tensor and batch it to avoid memory leaks
     data_tensor = torch.from_numpy(data).to(device)
     data_dl = DataLoader(
@@ -575,15 +694,31 @@ def decompress(model_path, input_path, model_name, config):
             out = model.decode(data_batch).to(device)
             # Converting back to numpyarray
             out = detacher(out)
+            if config.save_error_bounded_deltas:
+                if idx in error_bound_batch:
+                    # Error Bounded Deltas added to Decompressed output
+                    delta_idx = np.where(error_bound_batch == idx)
+                    deltas = error_bound_deltas[delta_idx][0]
+                    delta_index = error_bound_index[delta_idx][0]
+                    row_idx, col_idx = delta_index
+
+                    for i in range(len(row_idx)):
+                        out[row_idx[i]][col_idx[i]] -= deltas[i]
+                        deltas_added += 1
+
             if idx == 0:
                 decompressed = out
             else:
                 decompressed = np.concatenate((decompressed, out))
 
+    if config.save_error_bounded_deltas:
+        print("Total Deltas Added - ", deltas_added)
+
     if config.data_dimension == 2 and config.model_type == "dense":
         decompressed = decompressed.reshape(
-            (len(decompressed), number_of_columns, number_of_columns)
+            (len(decompressed), original_shape[1], original_shape[2])
         )
+
     return decompressed, names, normalization_features
 
 
