@@ -17,6 +17,102 @@ from torch import nn
 from torch.nn import functional as F
 
 
+import torch.utils.data
+from torch.nn import functional as F
+from torch.autograd import Function
+from ..modules import helper
+
+
+class LowerBound(Function):
+    @staticmethod
+    def forward(ctx, inputs, bound):
+        b = torch.ones(inputs.size(), device=inputs.device) * bound
+        b = b.to(inputs.device)
+        ctx.save_for_backward(inputs, b)
+        return torch.max(inputs, b)
+
+    @staticmethod
+    def backward(ctx, grad_output):
+        inputs, b = ctx.saved_tensors
+
+        pass_through_1 = inputs >= b
+        pass_through_2 = grad_output < 0
+
+        pass_through = pass_through_1 | pass_through_2
+        return pass_through.type(grad_output.dtype) * grad_output, None
+
+
+class GDN(nn.Module):
+    """Generalized divisive normalization layer.
+    y[i] = x[i] / sqrt(beta[i] + sum_j(gamma[j, i] * x[j]^2))
+    """
+
+    def __init__(
+        self,
+        ch,
+        inverse=False,
+        beta_min=1e-6,
+        gamma_init=0.1,
+        reparam_offset=2**-18,
+    ):
+        super(GDN, self).__init__()
+        self.inverse = inverse
+        self.beta_min = beta_min
+        self.gamma_init = gamma_init
+        self.device = helper.get_device()
+        self.reparam_offset = torch.tensor([reparam_offset], device=self.device)
+
+        self.build(ch)
+
+    def build(self, ch):
+        self.pedestal = self.reparam_offset**2
+        self.beta_bound = (self.beta_min + self.reparam_offset**2) ** 0.5
+        self.gamma_bound = self.reparam_offset
+
+        # Create beta param
+        beta = torch.sqrt(torch.ones(ch, device=self.device) + self.pedestal)
+        self.beta = nn.Parameter(beta)
+
+        # Create gamma param
+        eye = torch.eye(ch, device=self.device)
+        g = self.gamma_init * eye
+        g = g + self.pedestal
+        gamma = torch.sqrt(g)
+        self.gamma = nn.Parameter(gamma)
+
+    def forward(self, inputs):
+        unfold = False
+        if inputs.dim() == 5:
+            unfold = True
+            bs, ch, d, w, h = inputs.size()
+            inputs = inputs.view(bs, ch, d * w, h)
+
+        _, ch, _, _ = inputs.size()
+
+        # Beta bound and reparam
+        beta = LowerBound.apply(self.beta, self.beta_bound)
+        beta = beta**2 - self.pedestal
+
+        # Gamma bound and reparam
+        gamma = LowerBound.apply(self.gamma, self.gamma_bound)
+        gamma = gamma**2 - self.pedestal
+        gamma = gamma.view(ch, ch, 1, 1)
+
+        # Norm pool calc
+        norm_ = nn.functional.conv2d(inputs**2, gamma, beta)
+        norm_ = torch.sqrt(norm_)
+
+        # Apply norm
+        if self.inverse:
+            outputs = inputs * norm_
+        else:
+            outputs = inputs / norm_
+
+        if unfold:
+            outputs = outputs.view(bs, ch, d, w, h)
+        return outputs
+
+
 class AE(nn.Module):
     # This class is a modified version of the original class by George Dialektakis found at
     # https://github.com/Autoencoders-compression-anomaly/Deep-Autoencoders-Data-Compression-GSoC-2021
@@ -98,8 +194,6 @@ class CFD_dense_AE(nn.Module):
 
         self.activations = {}
 
-        n_features = n_features * n_features
-
         # encoder
         self.en1 = nn.Linear(n_features, 200, dtype=torch.float)
         self.en2 = nn.Linear(200, 100, dtype=torch.float)
@@ -224,7 +318,8 @@ class Conv_AE(nn.Module):
         super(Conv_AE, self).__init__(*args, **kwargs)
 
         self.q_z_mid_dim = 2000
-        self.q_z_output_dim = 72128
+        self.q_z_output_dim = 128
+        self.conv_op_shape = None
 
         # Encoder
 
@@ -242,6 +337,7 @@ class Conv_AE(nn.Module):
         )
         # Flatten
         self.flatten = nn.Flatten(start_dim=1)
+
         # Linear layers
         self.q_z_lin = nn.Sequential(
             nn.Linear(self.q_z_output_dim, self.q_z_mid_dim),
@@ -276,6 +372,9 @@ class Conv_AE(nn.Module):
     def encode(self, x):
         # Conv
         out = self.q_z_conv(x)
+        self.conv_op_shape = out.shape
+        self.q_z_output_dim = out.shape[1] * out.shape[2] * out.shape[3]
+
         # Flatten
         out = self.flatten(out)
         # Dense
@@ -286,7 +385,12 @@ class Conv_AE(nn.Module):
         # Dense
         out = self.p_x_lin(z)
         # Unflatten
-        out = out.view(out.size(0), 32, 49, 46)
+        out = out.view(
+            self.conv_op_shape[0],
+            self.conv_op_shape[1],
+            self.conv_op_shape[2],
+            self.conv_op_shape[3],
+        )
         # Conv transpose
         out = self.p_x_conv(out)
         return out
@@ -296,6 +400,12 @@ class Conv_AE(nn.Module):
         out = self.decode(z)
         return out
 
+    def get_final_layer_dims(self):
+        return self.conv_op_shape
+
+    def set_final_layer_dims(self, conv_op_shape):
+        self.conv_op_shape = conv_op_shape
+
 
 class FPGA_prototype_model(nn.Module):
     def __init__(self, n_features, z_dim, *args, **kwargs):
@@ -358,21 +468,31 @@ class Conv_AE_3D(nn.Module):
         super(Conv_AE_3D, self).__init__(*args, **kwargs)
 
         self.q_z_mid_dim = 2000
-        self.q_z_output_dim = 0  # Please change this to 4800000
-        self.compress_to_latent_space = False
+        self.q_z_output_dim = 612864
+        self.compress_to_latent_space = True
+        self.debug = False
 
         # Encoder
 
         # Conv Layers
         self.q_z_conv = nn.Sequential(
-            nn.Conv3d(1, 8, kernel_size=(1, 1, 1), stride=(1), padding=(1)),
+            nn.Conv3d(1, 4, kernel_size=(1, 3, 5), stride=(1), padding=(0)),
             # nn.BatchNorm2d(8),
             nn.ReLU(),
-            nn.Conv3d(8, 16, kernel_size=(3), stride=(1), padding=(1)),
+            nn.Conv3d(4, 8, kernel_size=(3), stride=(1), padding=(0)),
+            # nn.BatchNorm2d(16),
+            nn.ReLU(),
+            nn.Conv3d(8, 16, kernel_size=(3), stride=(1), padding=(0)),
             # nn.BatchNorm2d(16),
             nn.ReLU(),
             nn.Conv3d(16, 32, kernel_size=(3), stride=(1), padding=(0)),
-            # nn.BatchNorm2d(32),
+            # nn.BatchNorm2d(16),
+            nn.ReLU(),
+            nn.Conv3d(32, 64, kernel_size=(3), stride=(1), padding=(0)),
+            # nn.BatchNorm2d(16),
+            nn.ReLU(),
+            nn.Conv3d(64, 64, kernel_size=(1, 3, 3), stride=(1), padding=(0)),
+            # nn.BatchNorm2d(16),
             nn.ReLU(),
         )
         # Flatten
@@ -396,18 +516,26 @@ class Conv_AE_3D(nn.Module):
             # nn.BatchNorm1d(self.q_z_mid_dim),
             nn.Linear(self.q_z_mid_dim, self.q_z_output_dim),
             nn.ReLU(),
-            # nn.BatchNorm1d(self.q_z_output_dim),
         )
 
         # Conv Layers
         self.p_x_conv = nn.Sequential(
+            nn.ConvTranspose3d(64, 64, kernel_size=(1, 3, 3), stride=(1), padding=(0)),
+            nn.BatchNorm3d(64),
+            nn.ReLU(),
+            nn.ConvTranspose3d(64, 32, kernel_size=(3), stride=(1), padding=(0)),
+            nn.BatchNorm3d(32),
+            nn.ReLU(),
             nn.ConvTranspose3d(32, 16, kernel_size=(3), stride=(1), padding=(0)),
             nn.BatchNorm3d(16),
             nn.ReLU(),
-            nn.ConvTranspose3d(16, 8, kernel_size=(3), stride=(1), padding=(1)),
+            nn.ConvTranspose3d(16, 8, kernel_size=(3), stride=(1), padding=(0)),
             nn.BatchNorm3d(8),
             nn.ReLU(),
-            nn.ConvTranspose3d(8, 1, kernel_size=(1, 1, 1), stride=(1), padding=(1)),
+            nn.ConvTranspose3d(8, 4, kernel_size=(3), stride=(1), padding=(0)),
+            nn.BatchNorm3d(4),
+            nn.ReLU(),
+            nn.ConvTranspose3d(4, 1, kernel_size=(1, 3, 5), stride=(1), padding=(0)),
         )
 
     def encode(self, x):
@@ -424,12 +552,11 @@ class Conv_AE_3D(nn.Module):
         return out
 
     def decode(self, out):
-        # Unflatten
-        out = out.view(1, 32, 60, 50, 50)
-
         if self.compress_to_latent_space:
             out = self.p_x_lin(out)
 
+        out = out.view(4, 64, 7, 38, 36)
+
         # Conv transpose
         out = self.p_x_conv(out)
 
@@ -442,3 +569,97 @@ class Conv_AE_3D(nn.Module):
 
     def set_compress_to_latent_space(self, compress_to_latent_space):
         self.compress_to_latent_space = compress_to_latent_space
+
+
+class Conv_AE_GDN(nn.Module):
+    def __init__(self, n_features, z_dim, *args, **kwargs):
+        super(Conv_AE_GDN, self).__init__(*args, **kwargs)
+
+        self.q_z_mid_dim = 2000
+        self.q_z_output_dim = 128
+        self.conv_op_shape = None
+
+        # Encoder
+
+        # Conv Layers
+        self.q_z_conv = nn.Sequential(
+            nn.Conv2d(1, 8, kernel_size=(2, 5), stride=(1), padding=(1)),
+            # nn.BatchNorm2d(8),
+            GDN(8),
+            nn.Conv2d(8, 16, kernel_size=(3), stride=(1), padding=(1)),
+            nn.BatchNorm2d(16),
+            GDN(16),
+            nn.Conv2d(16, 32, kernel_size=(3), stride=(1), padding=(0)),
+            # nn.BatchNorm2d(32),
+            GDN(32),
+        )
+        # Flatten
+        self.flatten = nn.Flatten(start_dim=1)
+
+        # Linear layers
+        self.q_z_lin = nn.Sequential(
+            nn.Linear(self.q_z_output_dim, self.q_z_mid_dim),
+            nn.ReLU(),
+            # nn.BatchNorm1d(self.q_z_output_dim),
+            nn.Linear(self.q_z_mid_dim, z_dim),
+            nn.ReLU(),
+        )
+
+        # Decoder
+
+        # Linear layers
+        self.p_x_lin = nn.Sequential(
+            nn.Linear(z_dim, self.q_z_mid_dim),
+            nn.ReLU(),
+            # nn.BatchNorm1d(self.q_z_output_dim),
+            nn.Linear(self.q_z_mid_dim, self.q_z_output_dim),
+            nn.ReLU()
+            # nn.BatchNorm1d(42720)
+        )
+        # Conv Layers
+        self.p_x_conv = nn.Sequential(
+            nn.ConvTranspose2d(32, 16, kernel_size=(3), stride=(1), padding=(0)),
+            # nn.BatchNorm2d(16),
+            GDN(16, inverse=True),
+            nn.ConvTranspose2d(16, 8, kernel_size=(3), stride=(1), padding=(1)),
+            # nn.BatchNorm2d(8),
+            GDN(8, inverse=True),
+            nn.ConvTranspose2d(8, 1, kernel_size=(2, 5), stride=(1), padding=(1)),
+        )
+
+    def encode(self, x):
+        # Conv
+        out = self.q_z_conv(x)
+        self.conv_op_shape = out.shape
+        self.q_z_output_dim = out.shape[1] * out.shape[2] * out.shape[3]
+
+        # Flatten
+        out = self.flatten(out)
+        # Dense
+        out = self.q_z_lin(out)
+        return out
+
+    def decode(self, z):
+        # Dense
+        out = self.p_x_lin(z)
+        # Unflatten
+        out = out.view(
+            self.conv_op_shape[0],
+            self.conv_op_shape[1],
+            self.conv_op_shape[2],
+            self.conv_op_shape[3],
+        )
+        # Conv transpose
+        out = self.p_x_conv(out)
+        return out
+
+    def forward(self, x):
+        z = self.encode(x)
+        out = self.decode(z)
+        return out
+
+    def get_final_layer_dims(self):
+        return self.conv_op_shape
+
+    def set_final_layer_dims(self, conv_op_shape):
+        self.conv_op_shape = conv_op_shape
